# OB - Framework to program bots
#
# ob/irc.py (internet relay chat)
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 13-07-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

""" irc bot. """

__version__ = 1

## imports

from ob import Obj, locked
from ob.bot import Bot
from ob.cli import CLI
from ob.dcc import DCC
from ob.db import Db
from ob.event import Event 
from ob.fleet import Fleet
from ob.loader import Loader
from ob.handler import Handler
from ob.shell import defaults, Main
from ob.thr import Launcher
from ob.trace import get_exception
from ob.users import Users

import _thread
import logging
import ob
import os
import queue
import random
import re
import socket
import ssl
import textwrap
import time
import threading

## instances

db = Db()
fleet = Fleet()
launcher = Launcher()
users = Users()

## defines

class Cfg(Obj):

    pass

class Event(Event):

    pass

## classes

class IRC(Bot):

    cc = "!"

    def __init__(self):
        super().__init__()
        self.register("352", self.handler)
        self.register("366", self.handler)
        self.register("376", self.handle_connect)
        self.register("433", self.handler)
        self.register("ERROR", self.handle_error)
        self.register("MODE", self.handler)
        self.register("PING", self.handler)
        self.register("PONG", self.handler)
        self.register("PRIVMSG", self.handle_privmsg)
        self._connected = threading.Event()
        self._last = 0
        self.buffer = []
        self.cfg = Cfg(defaults).last()
        if self.cfg.channel:
            self.channels.append(self.cfg.channel)
        self.lastline = ""
        self.joined = []
        self.nrsend = 0
        self.nrconnect = 0

    def announce(self, txt):
        for channel in self.channels:
            self.say(channel, txt)

    def close(self):
        self.sock.shutdown(1)
        self.sock.close()
        self.fsock.close()

    def connect(self):
        self.cfg.blocking = True
        oldsock = None
        self.cfg.server = self.cfg.server or "localhost"
        self.cfg.port = self.cfg.port or 6667
        logging.warn("connect %s %s" % (self.cfg.server, self.cfg.port))
        if self.cfg.ipv6:
            oldsock = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        else:
            oldsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        oldsock.setblocking(self.cfg.blocking)
        oldsock.settimeout(60.0)
        oldsock.connect((self.cfg.server, int(self.cfg.port)))
        oldsock.setblocking(self.cfg.blocking)
        oldsock.settimeout(700.0)
        if "ssl" in self.cfg and self.cfg.ssl:
            self.sock = ssl.wrap_socket(oldsock)
        else:
            self.sock = oldsock
        self.fsock = self.sock.makefile("r")
        os.set_inheritable(self.sock.fileno(), os.O_RDWR)
        return True

    def command(self, cmd, *args, direct=False):
        if not args:
            self.raw("%s %s" % (cmd.upper(), args[0]), direct)
            return
        if len(args) == 1:
            self.raw("%s %s" % (cmd.upper(), args[0]), direct)
            return
        if len(args) >= 1:
            self.raw("%s %s :%s" % (cmd.upper(), args[0], " ".join(args[1:])), direct)
            return
        
    def dispatch(self, event):
        if event.command:
            func = self.get_handler(event.command)
            if func:
                func(event)
                return
        super().dispatch(event)

    def get_event(self, txt=""):
        if not txt and not self.buffer:
            self.some()
        line = txt or self.buffer.pop(0)
        return self.parsing(line.rstrip())

    def handler(self, event):
        cmd = event.command
        if cmd == "001":
            self.logon()
        elif cmd == "PING":
            self.pongcheck = True
            self.command("PONG", event.txt, direct=True)
        elif cmd == "PONG":
            self.pongcheck = False
        elif cmd == "352":
            args = event.arguments
            users.userhosts[args[5]] = args[2] + "@" + args[3]
        elif cmd == "366":
            self.joined.append(event.channel)
        elif cmd == "433":
            nick = event.target + "_"
            self.cfg.nick = nick
            self.command("NICK", nick)

    def handle_connect(self, event):
        if "servermodes" in self:
            self.raw("MODE %s %s" % (self.cfg.nick, self.cfg.servermodes), True)
        logging.error("connected to %s" % self.cfg.server)
        self.joinall()
        self._connected.set()

    @locked
    def handle_error(self, event):
        logging.error(event)
        self.stop()

    def handle_privmsg(self, event):
        self.callback(event)
        if event.txt.startswith("\001DCC"):
            try:
                dcc = DCC()
                dcc.encoding = "utf-8"
                self.launch(dcc.connect, event)
                return
            except ConnectionRefusedError:
                return
        elif event.txt.startswith("\001VERSION"):
            self.command("NOTICE", event.channel, "\001VERSION OB #%s - Framework to program bots - https://bitbucket.org/bthate/ob\001" % ob.__version__)
            return
        if event.txt and event.txt[0] != self.cc:
            return
        users.userhosts[event.nick] = event.origin
        if not users.allowed(event.origin, "USER"):
            return
        txt = event.txt[1:]
        e = Event(event, txt=txt)
        super().dispatch(e)
 
    def joinall(self):
        for channel in self.channels:
            self.command("JOIN", channel, direct=True)

    def logon(self, username="ob-irc", realname="ob-irc"):
        self.raw("NICK %s" % self.cfg.nick, direct=True)
        self.raw("USER %s %s %s :%s" % (username, self.cfg.server, self.cfg.server, realname), direct=True)

    def parsing(self, txt):
        rawstr = str(txt)
        obj = Event()
        obj.txt = ""
        obj.cc = self.cc
        obj.command = ""
        obj.arguments = []
        arguments = rawstr.split()
        obj.origin = arguments[0]
        if obj.origin.startswith(":"):
            obj.origin = obj.origin[1:]
            if len(arguments) > 1:
                obj.command = arguments[1]
            if len(arguments) > 2:
                txtlist = []
                adding = False
                for arg in arguments[2:]:
                    if arg.startswith(":"):
                        adding = True
                        txtlist.append(arg[1:])
                        continue
                    if adding:
                        txtlist.append(arg)
                    else:
                        obj.arguments.append(arg)
                    obj.txt = " ".join(txtlist)
        else:
            obj.cmd = obj.command = obj.origin
            obj.origin = self.cfg.server
        try:
            obj.nick, obj.origin = obj.origin.split("!")
        except:
            obj.nick = ""
        if obj.arguments:
            obj.target = obj.arguments[-1]
        else:
            obj.target = "" 
        if obj.target.startswith("#"):
            obj.channel = obj.target
        else:
            obj.channel = obj.nick
        if not obj.txt and len(arguments) == 1:
            obj.txt = arguments[1]
        if not obj.txt:
            obj.txt = rawstr.split(":")[-1]
        obj.args = obj.txt.split()
        obj.cb = obj.command
        obj.cmd = None
        obj.orig = repr(self)
        obj.rest = " ".join(obj.args)
        return obj

    @locked
    def raw(self, txt, direct=False):
        if self.stopped:
            return
        if not txt.startswith("PING") and not txt.startswith("PONG"):
            logging.warn("%s" % txt)
        if not txt.endswith("\r\n"):
            txt += "\r\n"
        txt = txt[:512]
        txt = bytes(txt, "utf-8")
        if not direct and (time.time() - self._last) < 3.0:
            time.sleep(1.0 * (self.nrsend % 10))
        self._last = time.time()
        self.nrsend += 1
        self.sock.send(txt)

    def say(self, channel, txt, type="normal", name=""):
        for txt in textwrap.wrap(txt, 500):
            self.command("PRIVMSG", channel, txt)

    def some(self, ssl=False, encoding="utf-8"):
        if ssl:
            inbytes = self.sock.read()
        else:
            inbytes = self.sock.recv(512)
        txt = str(inbytes, encoding)
        if txt == "":
            raise ConnectionResetError
        txt = bytes(txt, "utf-8").decode("unicode_escape")
        self.lastline += txt
        splitted = self.lastline.split("\r\n")
        for s in splitted[:-1]:
            self.buffer.append(s)
            if not s.startswith("PING") and not s.startswith("PONG"):
                logging.warn("%s" % s.strip())
        self.lastline = splitted[-1]

    def start(self):
        self.walk("ob.cbs.logger")
        self.cfg.nick = self.cfg.nick or "obbot"
        self.cfg.server = self.cfg.server or "localhost"
        self.cfg.channel = self.cfg.channel or "#obbot"
        self.cfg.save()
        self.connect()
        self.logon()
        super().start()
        return

    def stop(self):
        self.stopped = True
        super().stop()
        self.ready()
