# OB - Framework to program bots
#
# ob/shell.py
#
# Copyright 2017,2018 B.H.J Thate
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice don't have to be included.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
# THE SOFTWARE.
#
# 11-07-2018 As the creator of this file, I disclaim all rights on this file. 
#
# Bart Thate
# Heerhugowaard
# The Netherlands

"""
    utility functions.

"""

__version__ = 1

## imports

from ob import Obj, __version__

import atexit
import logging
import logging.handlers
import ob
import optparse
import os
import readline
import rlcompleter
import stat
import sys
import termios
import time

## defines

homedir = os.path.expanduser("~")
curdir = os.getcwd()
logdir = homedir + os.sep + ".oblogs"
dirmask = 0o700
filemask = 0o600
histfile = os.path.join(logdir, "history")
resume = {}
datefmt = '%H:%M:%S'

format_large = "%(asctime)-8s %(module)20s.%(lineno)-4s %(message)-50s"
format_source = "%(asctime)-8s %(message)-70s (%(module)s.%(lineno)s)"
format_time = "%(asctime)-8s %(message)-70s"
format_log = "%(message)s"
format_super = "%(asctime)-8s -=- %(message)-50s -=- (%(module)s.%(lineno)-10s) (%(threadName)-10s)" 
format_normal = "%(asctime)s %(message)s"
format_plain = "%(message)-0s"

defaults = Obj()
defaults.all = ""
defaults.channel = ""
defaults.daemon = False
defaults.kind = "main"
defaults.level = ""
defaults.modules = ""
defaults.init = ""
defaults.ipv6 = False
defaults.nick = ""
defaults.port = ""
defaults.server = ""
defaults.ssl = False
defaults.shell = False
defaults.verbose = False
defaults.workdir = ob.workdir
defaults.write = False

opts_main = [
    ('-a', '--all', 'store_true', False, 'all', 'load all modules.'),
    ('-c', '--connect', 'string', '', 'server', "server to connect to."),
    ("-d", "--daemon", 'store_true', False, "daemon", 'enable daemon mode.'),
    ('-i', '--init', 'string', '', 'init', "comma seperated init script names."),
    ('-j', '--join', 'string', '', 'channel', "channel/room to join."),
    ('-l', '--level', 'string', "error", 'level', "loglevel."),
    ('-m', '--modules', 'string', '', "modules", 'comma seperated list of modules.'),
    ("-n", "--nick", "string", "", "nick", "nick to use when joining a room."),
    ('-p', "--port", "string", "", "port", "server port to connect to."),
    ('-v', '--verbose', 'store_true', False, 'verbose', 'enable verbose mode.'),
    ('-w', '--workdir', 'string', ob.workdir, "workdir", "set working directory."),
    ('-u', '--user', 'string', '', 'user', 'jabber user (JID) to connect with.'),
    ('-z', '--shell', 'store_true', False, 'shell', 'enable shell.'),
    ('', '--ipv6', 'store_true', False, 'ipv6', 'enable ipv6.'),
    ('', '--license', 'store_true', False, 'license', 'show license.'),
    ('', '--resume', 'store_true', False, 'resume', 'flag set by the bot when rebooting.'),
    ('', '--wait', 'store_true', False, 'wait', 'wait for bot to finish.'),
    ("", "--password", 'store_true', False, "password", "password to connect with."),
    ("", "--prompted", 'store_true', False, "prompted", "use prompt."),
    ('', '--ssl', 'store_true', False, 'ssl', 'enable ssl')
] 

## classes

class Completer(rlcompleter.Completer):

    def __init__(self, options):

        self.options = sorted(options)

    def complete(self, text, state):
        if state == 0:
            if text:
                self.matches = [s for s in self.options if s and s.startswith(text)]
            else:
                self.matches = self.options[:]
        if state < len(self.options):
            return self.matches[state]
        else:
            return None

    def display(self, substitution, matches, depth):
        if matches:
            print("\t".join(matches))
            readline.redisplay()
    
class DumpHandler(logging.StreamHandler):

    def emit(self, record):
        pass

class Main(Obj):

    pass

## functions

def banner(name=""):
    return license()

def cdir(path):
    res = ""
    for p in path.split(os.sep):
        res += "%s%s" % (p, os.sep)
        padje = os.path.abspath(os.path.normpath(res))
        try:
            os.mkdir(padje)
        except (IsADirectoryError, NotADirectoryError, FileExistsError):
            pass
    check_permissions(padje)
    return True

def check_permissions(p, dirmask=dirmask, filemask=filemask):
    uid = os.getuid()
    gid = os.getgid()
    try:
        stats = os.stat(p)
    except FileNotFoundError:
        return
    except OSError:
        d, fn = os.path.split(p)
        cdir(d)
        stats = os.stat(d)
    if stats.st_uid != uid:
        os.chown(p, uid, gid)
    if os.path.isfile(p):
        mask = filemask
    else:
        mask = dirmask
    m = oct(stat.S_IMODE(stats.st_mode))
    if m != oct(mask):
        os.chmod(p, mask)

def close_history():
    readline.write_history_file(histfile)

def daemon():
    pid = os.fork()
    if pid != 0:
        os._exit(0)

def enable_history():
    if not os.path.exists(histfile):
        d, f = os.path.split(histfile)
        cdir(d)
        touch(histfile)
    readline.read_history_file(histfile)
    atexit.register(close_history)

def get_completer():
    return readline.get_completer()

def high(target, file_name):
    highest = 0
    for i in os.listdir(target):
        if file_name in i:
            try:
                seqnr = i.split('.')[-1]
            except IndexError:
                continue
            try:
                if int(seqnr) > highest:
                     highest = int(seqnr)
            except ValueError:
                pass
    return highest

def highest(target, filename):
    nr = high(target, filename)
    return "%s.%s" % (filename, nr+1)

def level(level, nostream=False):
    level = level.upper()
    logger = logging.getLogger("")
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    try:
        logger.setLevel(level)
    except ValueError:
        return
    formatter = logging.Formatter(format_plain, datefmt)
    if nostream:
        dhandler = DumpHandler()
        dhandler.propagate = False
        dhandler.setLevel(level)
        logger.addHandler(dhandler)
    else:
        ch = logging.StreamHandler()
        ch.propagate = False
        ch.setFormatter(formatter)
        ch.setLevel(level)
        logger.addHandler(ch)
    formatter2 = logging.Formatter(format_time, datefmt)
    filehandler = logging.handlers.TimedRotatingFileHandler(os.path.join(logdir, "log"), 'midnight')
    filehandler.propagate = False
    filehandler.setFormatter(formatter2)
    filehandler.setLevel(level)
    logger.addHandler(filehandler)
    return logger
    
def license():
    import ob
    return "".join(open(str(ob.__file__), "r").readlines()[:26])

def locatedir(path, match=""):
    for root, dirs, files in os.walk(path):
        for d in dirs:
            if root == path:
                continue
            if match and match not in d:
                continue
            yield root

def main():
    from ob.cli import CLI
    cfg = parse_cli("ob")
    level(cfg.level)
    cfg.save()
    if cfg.daemon:
        daemon()
    bot = CLI()
    if cfg.all:
        bot.walk("ob.cmds")
    for name in cfg.modules.split(","):
        if not name:
            continue
        bot.walk(name, True)
    for name in cfg.init.split(","):
        if not name:
            continue
        cfg.shell = True
        n = "ob.init.%s" % name
        bot.walk(n, True)
    bot.start()
    if cfg.shell:
        bot._prompted = True
        bot.prompt()
    if cfg.shell or cfg.daemon:
        bot.wait()
    else:
        bot.cmd(" ".join(cfg.args))

def make_opts(options, usage, version):
    parser = optparse.OptionParser(usage=usage, version=version)
    for option in options:
        optiontype, default, dest, helptype = option[2:]
        if "store" in optiontype:
            try:
                parser.add_option(option[0], option[1], action=optiontype, default=default, dest=dest, help=helptype)
            except Exception as ex:
                logging.error("%s option %s" % (str(ex), option))
                continue
        else:
            try:
                parser.add_option(option[0], option[1], type=optiontype, default=default, dest=dest, help=helptype)
            except Exception as ex:
                logging.error("^%s option %s" % (str(ex), option))
                continue
    args = parser.parse_args()
    return args

def parse_cli(name, options=None, version=__version__):
    usage = "%s [options]" % name
    ver = "%s #%s" % (name.upper(), version)
    if not options:
        options = opts_main
    opt, arguments = make_opts(options, usage, ver)
    cfg = Main(defaults)
    cfg.upgrade(vars(opt))
    cfg.args = arguments
    if cfg.workdir:
        import ob
        ob.workdir = cfg.workdir
    return cfg

def reset():
    close_history()
    if "old" in resume:
        try:
            termreset(resume["fd"], resume["old"])
        except:
            pass

def set_completer(optionlist):
    completer = Completer(optionlist)
    readline.set_completer_delims(' ')
    readline.parse_and_bind("tab: complete")
    readline.set_completion_display_matches_hook(completer.display)
    readline.set_completer(completer.complete)
    atexit.register(lambda: readline.set_completer(None))

def startup():
    global resume
    enable_history()
    resume["fd"] = sys.stdin.fileno()
    resume["old"] = termsetup(sys.stdin.fileno())
    atexit.register(reset)

def termreset(fd, old):
    termios.tcsetattr(fd, termios.TCSADRAIN, old)

def termsetup(fd):
    old = termios.tcgetattr(fd)
    return old

def touch(fname):
    try:
        fd = os.open(fname, os.O_RDONLY | os.O_CREAT)
        os.close(fd)
    except TypeError:
        pass
